; $Id: real.scm,v 1.2 2006/01/09 21:53:31 schwicht Exp $

; Constructive Reals
; ==================

; To work with reals, we use a predicate constant Cauchy which takes
; two arguments, a sequence of rationals and a modulus.

; We introduce Cauchy as an inductively defined predicate (which may
; in this case also be called a record).

(add-ids (list (list "Cauchy"
		     (make-arity (py "pos=>rat") (py "pos=>pos"))))
	 '("all as,M.(all k,n,m.M k<=n -> M k<=m ->
                                abs(as n-as m)<=(1#exp 2 k)) ->
                     Cauchy as M" "CauchyIntro"))

(define idpc
  (idpredconst-name-and-types-and-cterms-to-idpredconst "Cauchy" '() '()))

; There are no types, since the clauses do not contain type variables,
; and no cterms, since the clauses do not contain parameter predicate
; variables.

(define aconst0 (number-and-idpredconst-to-intro-aconst 0 idpc))
(pp (aconst-to-formula aconst0))

; Similarly, we introduce a predicate constant Mon, taking a sequence
; of positive numbers as argument, to express monotonicity.

(add-ids (list (list "Mon" (make-arity (py "pos=>pos"))))
	 '("all M.(all n,m.n<=m -> M n<=M m) -> Mon M" "MonIntro"))

; The meaning of the predicate constants is determined by introduction
; and elimination axioms:

; (set-goal
;  (pf "all as,M.(all k,n,m.M k<=n -> M k<=m ->
;                           abs(as n-as m)<=(1#exp 2 k)) ->
;                Cauchy as M"))
; (strip)
; (intro 0)
; (use 1)
; (save "CauchyIntro")

(set-goal
 (pf "all as,M.Cauchy as M ->
               all k,n,m.M k<=n -> M k<=m ->
                         abs(as n-as m)<=(1#exp 2 k)"))
(assume "as" "M")
(elim)
(search)
(save "CauchyElim")

; (set-goal (pf "all M.(all n,m.n<=m -> M n<=M m) -> Mon M"))
; (strip)
; (intro 0)
; (use 1)
; (save "MonIntro")

(set-goal (pf "all M.Mon M -> all m,n.m<=n -> M m<=M n"))
(assume "M")
(elim)
(search)
(save "MonElim")

; We introduce Real as an inductively defined predicate (which in this
; case may also be called a record).  Then we can prove theorems:

; RealIntro: all x.Cauchy(x seq)(x mod) -> Mon(x mod) -> Real x
; RealElim1: all as,M.Real(RealConstr as M) -> Cauchy as M
; RealElim2: all as,M.Real(RealConstr as M) -> Mon M

; Alternative formulation (worse, since usability is restricted)
; RealIntro: all as,M.Cauchy as M -> Mon M -> Real RealConstr as M, 
; RealElim1: all x.Real x -> Cauchy(x seq)(x mod)
; RealElim2: all x.Real x -> Mon(x mod)

(add-ids
 (list (list "Real" (make-arity (py "real"))))
 '("all x.Cauchy(x seq)(x mod) -> Mon(x mod) -> Real x" "RealIntro"))

; (set-goal (pf "all x.Cauchy(x seq)(x mod) -> Mon(x mod) -> Real x"))
; (strip)
; (intro 0)
; (auto)
; (save "RealIntro")

(set-goal (pf "all x.Real x -> Cauchy(x seq)(x mod)"))
(assume "x")
(elim)
(auto)
(save "RealElim1")

(set-goal (pf "all x.Real x -> Mon(x mod)"))
(assume "x")
(elim)
(auto)
(save "RealElim2")

; The following variants will be more useful, because their heads will
; be more often of the form of a given goal.

(set-goal (pf "all as,M.Real(RealConstr as M) -> Cauchy as M"))
(strip)
(use-with "RealElim1" (pt "RealConstr as M") 1)
(save "RealElimVariant1")

(set-goal (pf "all as,M.Real(RealConstr as M) -> Mon M"))
(strip)
(use-with "RealElim2" (pt "RealConstr as M") 1)
(save "RealElimVariant2")

(set-goal (pf "all x,y.Real x -> Real y -> x+y===y+x"))
(strip)
(ord-field-simp-bwd)
(use 1)
(use 2)

(set-goal (pf "all x,y,k,l.Real x -> Real y -> 
               RealLt 0(abs x)k -> RealLt 0(abs y)l ->
               0<<=x*y*((IntN 1#1)*x+y) ->
               1/y<<=1/x"))
(strip)
(ord-field-simp-bwd)
(auto)

; To prove transitivity of equality, we need a characterization of
; equality.

(add-global-assumption
 "RatAbsLe3"
 (pf "all a,b,c,d.abs(a-b)<=abs(a-c)+abs(c-d)+abs(d-b)"))

(add-global-assumption
 "RatPlusLe3"
 (pf "all a1,a2,b1,b2,c1,c2.a1<=a2 -> b1<=b2 -> c1<=c2 ->
                            a1+b1+c1<=a2+b2+c2"))

(add-global-assumption "PosLeTrans" (pf "all n,m,k.n<=m -> m<=k -> n<=k"))
(add-global-assumption "RatLeTrans" (pf "all a,b,c.a<=b -> b<=c -> a<=c"))

(add-global-assumption "PosMaxUB1" (pf "all n,m.n<=n max m"))
(add-global-assumption "PosMaxUB2" (pf "all n,m.m<=n max m"))

; We now prove one half of the characterization of equality for reals,
; to be called "RealEqChar1"

; In the following proof for the first time we benefit from omission
; of the computation rule exp n(S m) -> n*exp n m.

; (display-program-constants "PosExp")

(set-goal
 (pf "allnc as all M allnc bs all N.Cauchy as M -> Cauchy bs N ->
      RealConstr as M===RealConstr bs N -> 
      all k ex q all n.q<=n -> abs(as n-bs n)<=(1#exp 2 k)"))
(strip)
(ex-intro (pt "M(k+2)max N(k+2)"))
(strip)
(use "RatLeTrans" (pt "(1#exp 2(k+2))+(1#exp 2(k+1))+(1#exp 2(k+2))"))

; ?_5: abs(as n-bs n)<=(1#exp 2(k+2))+(1#exp 2(k+1))+(1#exp 2(k+2))
(use "RatLeTrans" (pt "abs(as n-as(M(k+2)))+
                        abs(as(M(k+2))-bs(N(k+2)))+
                        abs(bs(N(k+2))-bs n)"))

; ?_7: abs(as n-bs n)<=abs(as n-as(M(k+2)))+abs(as(M(k+2))-bs(N(k+2)))+abs(bs(N(k+2))-bs n)
(use "RatAbsLe3")

; ?_8: abs(as n-as(M(k+2)))+abs(as(M(k+2))-bs(N(k+2)))+abs(bs(N(k+2))-bs n)<=(1#exp 2(k+2))+(1#exp 2(k+1))+(1#exp 2(k+2))
(use "RatPlusLe3")

; ?_9: abs(as n-as(M(k+2)))<=(1#exp 2(k+2))
(use "CauchyElim" (pt "M"))
(use 1)
(use "PosLeTrans" (pt "(M(k+2))max(N(k+2))"))
(use "PosMaxUB1")
(use 4)
(use "Truth-Axiom")

; ?_10: abs(as(M(k+2))-bs(N(k+2)))<=(1#exp 2(k+1))
(use "RealEqElimVariant")
(use 3)

; ?_11: abs(bs(N(k+2))-bs n)<=(1#exp 2(k+2))
(use "CauchyElim" (pt "N"))
(use 2)
(prop)
(use "PosLeTrans" (pt "(M(k+2))max(N(k+2))"))
(use "PosMaxUB2")
(use 4)

; ?_6: (1#exp 2(k+2))+(1#exp 2(k+1))+(1#exp 2(k+2))<=(1#exp 2 k)
(ord-field-simp-bwd)

(save "RealEqChar1")

; (dp)
; (dpe)

(define RealEqChar1-neterm
  (nt (proof-to-extracted-term (theorem-name-to-proof "RealEqChar1"))))

(pp RealEqChar1-neterm)
; "[M0,M1,n2]M0(S(S n2))max M1(S(S n2))"

; (define sproof (proof-to-soundness-proof (current-proof)))
; (cdp sproof)


; We now aim at the converse.  For this we need two auxiliary lemmata:

(add-global-assumption "RealEqChar2Aux1" (pf "all n,m.n*exp 2 m<m -> F"))
(add-global-assumption "RealEqChar2Aux2" (pf "all n,m.n*exp 2 m=m -> F"))

(set-goal (pf "all a.all k 0<=a+(1#exp 2 k) -> 0<=a"))
(cases)
(cases)

; ?_3: all n559,n555.all k 0<=(n559#n555)+(1#exp 2 k) -> 0<=(n559#n555)
(strip)
(ng)
(use "Truth-Axiom")

; ?_4: all n555.all k 0<=(0#n555)+(1#exp 2 k) -> 0<=(0#n555)
(strip)
(ng)
(use "Truth-Axiom")

; ?_5: all n558,n555.all k 0<=(IntN n558#n555)+(1#exp 2 k) -> 0<=(IntN n558#n555)
(assume "n" "m" "Est")
(cut (pf "0<=(IntN n#m)+(1#exp 2 m)"))
(drop "Est")
(ng)
(cut (pf "n*exp 2 m=m -> F"))
(assume "EqHyp")
(simp "EqHyp")
(cut (pf "n*exp 2 m<m -> F"))
(assume "LtHyp")
(simp "LtHyp")
(ng)
(prop)
(use "RealEqChar2Aux1")
(use "RealEqChar2Aux2")
(use "Est")

(save "RatLeCrit")

; We need this is a slightly more general form

(add-global-assumption "RatLeAux1" (pf "all a,b.0<=b-a -> a<=b"))
(add-global-assumption "RatLeAux2" (pf "all a,b,c.a<=b+c -> 0<=b-a+c"))

(set-goal (pf "all a,b.all k a<=b+(1#exp 2 k) -> a<=b"))
(assume "a" "b" "Est")
(use "RatLeAux1")
(use "RatLeCrit")
(assume "k")
(use "RatLeAux2")
(use "Est")

(save "RatLeCritGen")

; We now prove the other half of the characterization of equality for
; reals, to be called "RealEqChar2".

(set-goal
 (pf "all as,M,bs,N.Cauchy as M -> Cauchy bs N ->
      (all k ex n0 all n.n0<=n -> abs(as n-bs n)<=(1#exp 2 k)) ->
      RealConstr as M===RealConstr bs N"))
(strip)
(use "RealEqIntro")
(assume "k")
(use "RatLeCritGen")
(assume "l")
(ng)

; ?_7 abs(as(M(S k))-bs(N(S k)))<=(exp 2 l+exp 2 k#exp 2 k*exp 2 l)
(inst-with-to 3 (pt "l") "InstExHyp")
(drop 3)
(by-assume-with "InstExHyp" "n0" "InstExHypKernel")

; We now want to use n as an abbreviation for ((M(k+1))max(N(k+1)))max n0
; Hence we introduce via cut the formula all n.n=term -> goal
(cut (pf "all n.n=((M(k+1))max(N(k+1)))max n0 ->
          abs(as(M(k+1))-bs(N(k+1)))<=
          (1#exp 2 k)+(1#exp 2 l)"))
(assume "allHyp")
(use "allHyp" (pt "((M(k+1))max(N(k+1)))max n0"))
(prop)
(assume "n" "nDef")
(inst-with-to "InstExHypKernel" (pt "n") "Mid")
(drop  "InstExHypKernel")
(use "RatLeTrans" (pt "(1#exp 2(k+1))+
                        (1#exp 2 l)+(1#exp 2(k+1))"))

; ?_22: abs(as(M(k+1))-bs(N(k+1)))<=(1#exp 2(k+1))+(1#exp 2 l)+(1#exp 2(k+1))
(use "RatLeTrans" (pt "abs(as(M(k+1))-as n)+abs(as n-bs n)+
                        abs(bs n-bs(N(k+1)))"))

; ?_24: abs(as(M(k+1))-bs(N(k+1)))<=abs(as(M(k+1))-as n)+abs(as n-bs n)+abs(bs n-bs(N(k+1)))
(use "RatAbsLe3")

; ?_25: abs(as(M(k+1))-as n)+abs(as n-bs n)+abs(bs n-bs(N(k+1)))<=(1#exp 2(k+1))+(1#exp 2 l)+(1#exp 2(k+1))
(use "RatPlusLe3")

; ?_26: abs(as(M(k+1))-as n)<=(1#exp 2(k+1))
(use "CauchyElim" (pt "M"))
(use 1)
(prop)
(simp "nDef")
(use "PosLeTrans" (pt "PosMax(M(k+1))(N(k+1))"))
(use "PosMaxUB1")
(use "PosMaxUB1")
(use "Mid")
(simp "nDef")
(use "PosMaxUB2")

; ?_28: abs(bs n-bs(N(k+1)))<=(1#exp 2(k+1))
(use "CauchyElim" (pt "N"))
(use 2)
(use "PosLeTrans" (pt "PosMax(M(k+1))(N(k+1))"))
(use "PosMaxUB2")
(simp "nDef")
(use "PosMaxUB1")
(use "Truth-Axiom")

; ?_23: (1#exp 2(k+1))+(1#exp 2 l)+(1#exp 2(k+1))<=(1#exp 2 k)+(1#exp 2 l)
(ord-field-simp-bwd)

(save "RealEqChar2")


; The sum of two Cauchy sequences is a Cauchy sequence:

(add-global-assumption
 "RatAbsLe2"
 (pf "all a,b,c.abs(a-b)<=abs(a-c)+abs(c-b)"))

(add-global-assumption
 "RatPlusLe2"
 (pf "all a1,a2,b1,b2.a1<=a2 -> b1<=b2 -> a1+b1<=a2+b2"))

(add-global-assumption "RatTimesAbs" (pf "all a,b.abs(a*b)==abs a*abs b"))
(add-global-assumption "RatTriang2" (pf "all a,b.abs(a+b)<=abs a+abs b"))
(add-global-assumption "RatEqLe" (pf "all a,b,c.a==b -> b<=c -> a<=c"))
(add-global-assumption "RatLeEq" (pf "all a,b,c.a<=b -> b==c -> a<=c"))
(add-global-assumption "RatAbsCompat" (pf "all a,b.a==b -> abs a==abs b"))

(set-goal (pf "Cauchy as M -> Cauchy bs N -> 
               Cauchy ([n]as n+bs n) ([k](M(k+1))max(N(k+1)))"))
(strip)
(use "CauchyIntro")
(strip)
(ng)

; ?_5: abs(as n+bs n-(as m+bs m))<=(1#exp 2 k)
(use "RatLeTrans" (pt "abs(as n-as m)+abs(bs n-bs m)"))

; ?_6: abs(as n+bs n-(as m+bs m))<=abs(as n-as m)+abs(bs n-bs m)
(use "RatEqLe" (pt "abs(as n-as m+(bs n-bs m))"))
(use "RatAbsCompat")

; ?_10: as n+bs n-(as m+bs m)==as n-as m+(bs n-bs m)
(ord-field-simp-bwd)

; ?_9: abs(as n-as m+(bs n-bs m))<=abs(as n-as m)+abs(bs n-bs m)
(use "RatTriang2")

; ?_7: abs(as n-as m)+abs(bs n-bs m)<=(1#exp 2 k)
(use "RatLeTrans" (pt "(1#exp 2(k+1))+(1#exp 2(k+1))"))

; ?_11: abs(as n-as m)+abs(bs n-bs m)<=(1#exp 2(k+1))+(1#exp 2(k+1))
(use "RatPlusLe2")

; ?_13: abs(as n-as m)<=(1#exp 2(k+1))
(use "CauchyElim" (pt "M"))

(use 1)
(use "PosLeTrans" (pt "(M(k+1))max(N(k+1))"))
(use "PosMaxUB1")
(ng)
(use 3)
(use "PosLeTrans" (pt "(M(k+1))max(N(k+1))"))
(use "PosMaxUB1")
(use 4)

; ?_14: abs(bs n-bs m)<=(1#exp 2(k+1))
(use "CauchyElim" (pt "N"))
(use 2)
(use "PosLeTrans" (pt "(M(k+1))max(N(k+1))"))
(use "PosMaxUB2")
(use 3)
(use "PosLeTrans" (pt "(M(k+1))max(N(k+1))"))
(use "PosMaxUB2")
(use 4)

; ?_12: (1#exp 2(k+1))+(1#exp 2(k+1))<=(1#exp 2 k)
(ord-field-simp-bwd)

(save "CauchyPlus")


; Every real has an upper bound, that is the reals are Archimedian ordered.

(add-global-assumption "PosNotLeToLt" (pf "all n,m.(n<=m -> F) -> m<n"))
(add-global-assumption
 "PosLtSZeroHalf" (pf "all n,m.SZero(PosHalf n)<m -> m<=n -> m=n"))

(set-goal (pf "all m,n.m<=n -> 
                       (m<=PosHalf n -> Pvar^) -> 
                       (PosHalf n<m -> m<=SZero(PosHalf n) -> Pvar^) -> 
                       (m=n -> Pvar^) -> Pvar^"))
(strip)
(cases (pt "m<=PosHalf n"))
(use 2)
(assume "m<=PosHalf n -> F")
(cases (pt "m<=SZero(PosHalf n)"))
(use 3)
(use "PosNotLeToLt")
(use "m<=PosHalf n -> F")
(assume "m<=SZero(PosHalf n) -> F")
(use 4)
(use "PosLtSZeroHalf")
(use "PosNotLeToLt")
(use "m<=SZero(PosHalf n) -> F")
(use 1)
(save "PosLeHalf")

(define PosLeHalf-neterm
  (nt (proof-to-extracted-term (theorem-name-to-proof "PosLeHalf"))))

(pp PosLeHalf-neterm)
"[n0,n1,(alpha3)_2,(alpha3)_3,(alpha3)_4]
 [if (n0<=PosHalf n1)
   (alpha3)_2
   [if (n0<=SZero(PosHalf n1)) (alpha3)_3 (alpha3)_4]]"

(add-global-assumption "RatMaxUB1" (pf "all a,b.a<=a max b"))
(add-global-assumption "RatMaxUB2" (pf "all a,b.b<=a max b"))

; We prove an induction scheme for pos based on PosHalf

(add-pvar-name "Q" (make-arity (py "pos")))

(set-goal (pf "Q^ 1 -> (all m.Q^(PosHalf m) -> Q^m) -> all n Q^n"))
(assume "Base" "Step")
(ind)
(use "Base")
(assume "n" "IHn")
(use "Step")
(use "IHn")
(assume "n" "IHn")
(use "Step")
(use "IHn")
(save "PosHalfInd")

(pp (nt (proof-to-extracted-term (current-proof))))
"[(alpha22)_0,(pos=>alpha22=>alpha22)_1]
 (Rec pos=>alpha22)(alpha22)_0([n3](pos=>alpha22=>alpha22)_1(SZero n3))
  ([n3](pos=>alpha22=>alpha22)_1(SOne n3))"

(set-goal (pf "all n,as ex a.all m.m<=n -> as m<=a"))
(use-with "PosHalfInd"
	  (make-cterm (pv "n") (pf "all as ex a.all m.m<=n -> as m<=a"))
	  "?" "?")

; Base
(strip)
(ex-intro (pt "as 1"))
(cases)
(prop)
(strip)
(prop)
(strip)
(prop)

; Step
(assume "n" "IHn" "as")
(inst-with-to "IHn" (pt "as") "IHlft")
(inst-with-to "IHn" (pt "[k]as(PosHalf n+k)") "IHrht")
(drop "IHn")
(by-assume-with "IHlft" "a0" "InstIHlft")
(by-assume-with "IHrht" "a1" "InstIHrht")
(ex-intro (pt "(a0 max a1)max(as n)"))
(assume "m" "m<=n")
(use "PosLeHalf" (pt "m") (pt "n"))
(use "m<=n")
(assume "m<=PosHalf n")
(drop "m<=n")
(use "RatLeTrans" (pt "a0"))
(use-with "InstIHlft" (pt "m") "m<=PosHalf n")
(use "RatLeTrans" (pt "a0 max a1"))
(use "RatMaxUB1")
(use "RatMaxUB1")

; ?_27: PosHalf n<m -> m<=SZero(PosHalf n) -> as m<=a0 max a1 max as n
(assume "PosHalf n<m" "m<=SZero(PosHalf n)")
(drop "m<=n")
(use "RatLeTrans" (pt "a1"))
(use "RatEqLe" (pt "as(PosHalf n+(m--PosHalf n))"))
(cut (pf "PosHalf n+(m--PosHalf n)=m"))
(assume "PosHalf n+(m--PosHalf n)=m")
(simp "PosHalf n+(m--PosHalf n)=m")
(use "Truth-Axiom")

(add-global-assumption "PosPlusMinus" (pf "all n,m.n<m -> n+(m--n)=m"))

(use "PosPlusMinus")
(use "PosHalf n<m")
(use "InstIHrht")

(add-global-assumption
 "PosMinusBoundZero" (pf "all n,m.n<m -> m<=SZero n -> m--n<=n"))

(use "PosMinusBoundZero")
(use "PosHalf n<m")
(use "m<=SZero(PosHalf n)")
(use "RatLeTrans" (pt "a0 max a1"))
(use "RatMaxUB2")
(use "RatMaxUB1")

; ?_28: m=n -> as m<=a0 max a1 max as n
(assume "m=n")
(simp "m=n")
(use "RatMaxUB2")

(save "RatSeqFinBound")

(add-var-name "G" (py "(pos=>rat)=>rat"))
	      
(pp (nt (proof-to-extracted-term
	 (expand-theorems (theorem-name-to-proof "RatSeqFinBound")))))

"(Rec pos=>(pos=>rat)=>rat)([as2]as2 1)
([n2,G3,as4]G3 as4 max G3([n5]as4(n2+n5))max as4(SZero n2))
([n2,G3,as4]G3 as4 max G3([n5]as4(n2+n5))max as4(SOne n2))"

(set-goal (pf "all as,M.Cauchy as M -> ex a all n as n<=a"))
(assume "as" "M" "Cauchy as M")
(cut (pf "ex a all m.m<M 1 -> as m<=a"))
(assume "ExHyp")
(by-assume-with "ExHyp" "a" "FinBound")
(ex-intro (pt "a max(as(M 1))+1"))

; ?_9: all n as n<=a max as(M 1)+1
(assume "n")
(cases (pt "n<M 1"))
(assume "n<M 1")
(use "RatLeTrans" (pt "a"))
(use "FinBound")
(use "n<M 1")
(use "RatLeTrans" (pt "a max(as(M 1))"))
(use "RatMaxUB1")
(use "Truth-Axiom")

; ?_12: (n<M 1 -> F) -> as n<=a max as(M 1)+1
(assume "n<M 1 -> F")
(use "RatLeTrans" (pt "as(M 1)+(as n-as(M 1))"))
(ord-field-simp-bwd)

; ?_21: as(M 1)+(as n-as(M 1))<=a max as(M 1)+1
(use "RatPlusLe2")

; ?_22: as(M 1)<=a max as(M 1)
(use "RatMaxUB2")

; ?_23: as n-as(M 1)<=1
(use "RatLeTrans" (pt "abs(as n-as(M 1))"))
(add-global-assumption "RatAbsUB1" (pf "all a a<=abs a"))
(use "RatAbsUB1")

; ?_27: abs(as n-as(M 1))<=1
(use "RatLeTrans" (pt "IntP 1#exp 2 1"))
(use "CauchyElim" (pt "M"))
(use "Cauchy as M")

(add-global-assumption "PosNotLtToLe" (pf "all n,m.(n<m -> F) -> m<=n"))

(use "PosNotLtToLe")
(use "n<M 1 -> F")
(use "Truth-Axiom")

(use "Truth-Axiom")

; ?_4: ex a all m.m<M 1 -> as m<=a
(inst-with-to "RatSeqFinBound" (pt "M 1") "H1")
(inst-with-to "H1" (pt "as") "ExFinBound")
(by-assume-with "ExFinBound" "a" "ExFinBoundKernel")
(ex-intro (pt "a"))
(assume "m" "m<M 1")
(use "ExFinBoundKernel")

(add-global-assumption "PosLtToLe" (pf "all n,m.n<m -> n<=m"))

(use "PosLtToLe")
(use "m<M 1")

(save "RealBound")

; (pp (nt (proof-to-extracted-term (theorem-name-to-proof "RealBound"))))
; "[as0,M1]cRatSeqFinBound(M1 1)as0 max as0(M1 1)+1"

; (pp (nt (proof-to-extracted-term (expand-theorems (theorem-name-to-proof "RealBound")))))

"[as0,M1]
 (Rec pos=>(pos=>rat)=>rat)([as2]as2 1)
 ([n2,G3,as4]G3 as4 max G3([n5]as4(n2+n5))max as4(SZero n2))
 ([n2,G3,as4]G3 as4 max G3([n5]as4(n2+n5))max as4(SOne n2))
 (M1 1)
 as0 max 
 as0(M1 1)+
 1"


; The product of two Cauchy sequences is a Cauchy sequence:

(set-goal
 (pf "all as,M,bs,N,k1,k2.Cauchy as M -> Cauchy bs N -> Mon M -> Mon N ->
      all n abs(as n)<=exp 2 k1 -> all n abs(bs n)<=exp 2 k2 -> 
      Cauchy([n]as n*bs n)([k]M(k+1+k2)max N(k+1+k1))"))
(strip)
(use "CauchyIntro")
(strip)
(ng)
(use "RatEqLe" (pt "abs(as n*(bs n-bs m)+(as n-as m)*bs m)"))
(use "RatAbsCompat")
(ord-field-simp-bwd)

; ?_7: abs(as n*(bs n-bs m)+(as n-as m)*bs m)<=(1#exp 2 k)
(use "RatLeTrans"
     (pt "abs(as n*(bs n-bs m))+abs((as n-as m)*bs m)"))

; ?_9: abs(as n*(bs n-bs m)+(as n-as m)*bs m)<=abs(as n*(bs n-bs m))+abs((as n-as m)*bs m)
(use "RatTriang2")

; ?_10: abs(as n*(bs n-bs m))+abs((as n-as m)*bs m)<=(1#exp 2 k)
(use "RatLeTrans" (pt "(1#exp 2(k+1))+(1#exp 2(k+1))"))

; ?_11: abs(as n*(bs n-bs m))+abs((as n-as m)*bs m)<=(1#exp 2(k+1))+(1#exp 2(k+1))
(use "RatPlusLe2")

; ?_13: abs(as n*(bs n-bs m))<=(1#exp 2(k+1))
(use "RatEqLe" (pt "abs(as n)*abs(bs n-bs m)"))

; ?_15: abs(as n*(bs n-bs m))==abs(as n)*abs(bs n-bs m)
(use "RatTimesAbs")

; ?_16: abs(as n)*abs(bs n-bs m)<=(1#exp 2(k+1))
(use "RatLeTrans" (pt "(exp 2 k1#1)*(1#exp 2(k+k1+1))"))

; ?_17: abs(as n)*abs(bs n-bs m)<=exp 2 k1*(1#exp 2(k+k1+1))

(add-global-assumption
 "RatLeAbsTimes"
 (pf "all a,b,a1,b1.abs a<=a1 -> abs b<=b1 -> abs a*abs b<=a1*b1"))

(use "RatLeAbsTimes")
(use 5)

; ?_20: abs(bs n-bs m)<=(1#exp 2(k+k1+1))
(use "CauchyElim" (pt "N"))
(auto 1 '("PosMaxUB2" 1) '("PosLeTrans" 1))
(ord-field-simp-bwd)

; ?_14: abs((as n-as m)*bs m)<=(1#exp 2(k+1))
(use "RatEqLe" (pt "abs(as n-as m)*abs(bs m)"))
(use "RatTimesAbs")

; ?_25: abs(as n-as m)*abs(bs m)<=(1#exp 2(k+1))
(use "RatLeTrans" (pt "(1#exp 2(k+k2+1))*(exp 2 k2#1)"))

; ?_26: abs(as n-as m)*abs(bs m)<=(1#exp 2(k+k2+1))*exp 2 k2
(use "RatLeAbsTimes")

; ?_28: abs(as n-as m)<=(1#exp 2(k+k2+1))
(use "CauchyElim" (pt "M"))
(auto 1 '("PosMaxUB1" 1) '("PosLeTrans" 1))

; ?_27: (1#exp 2(k+k2+1))*exp 2 k2<=(1#exp 2(k+1))
(ord-field-simp-bwd)

; ?_12: (1#exp 2(k+1))+(1#exp 2(k+1))<=(1#exp 2 k)
(ord-field-simp-bwd)

(save "CauchyTimes")


; The inverse of a positive real is a real.

(add-global-assumption "RatLtLe" (pf "all a,b,c.a<b -> b<=c -> a<c"))

(add-global-assumption "RatInv" (pf "all n,a.(1#n)<=a -> abs(1/a)<=n"))

(set-goal (pf "all as,M,l.Cauchy as M -> Mon M ->
               all n 0<as n ->
               (all n.M(l+1)<=n -> (1#exp 2(l+1))<=as n) ->
               Cauchy([n]1/(as n))([k]M(2*(l+1)+k))"))
(assume "as" "M" "l" "Cauchy as M" "Mon M" "as-is-posseq" "as-pos-l")
(use "CauchyIntro")
(assume "k" "n" "m" "M(2*(l+1)+k)<=n" "M(2*(l+1)+k)<=m")
(ng)

; As auxiliary lemmata we provide M(l+1)<=m and M(l+1)<=n
(cut (pf "M(l+1)<=n"))
(assume "M(l+1)<=n")
(cut (pf "M(l+1)<=m"))
(assume "M(l+1)<=m")

; ?_11: abs(1/as n-1/as m)<=(1#exp 2 k)
(use "RatEqLe"
     (pt "abs((1/as n)*(1/as m)*(as m-as n))"))

; ?_12: abs(1/as n-1/as m)==abs(1/as n*(1/as m)*(as m-as n))
(use "RatAbsCompat")

; ?_14: 1/as n-1/as m==1/as n*(1/as m)*(as m-as n)
(ord-field-simp-bwd)

; ?_15: 0<abs(as n)
(use "RatLtLe" (pt "as n"))
(use "as-is-posseq")
(use "RatAbsUB1")

; ?_25: 0<abs(as m)
(use "RatLtLe" (pt "as m"))
(use "as-is-posseq")
(use "RatAbsUB1")

; ?_13: abs(1/as n*(1/as m)*(as m-as n))<=(1#exp 2 k)
(use "RatLeTrans" (pt "(exp 2(l+1)#1)*(exp 2(l+1)#1)*
                        (1#exp 2(2*(l+1)+k))"))

; ?_21: abs(1/as n*(1/as m)*(as m-as n))<=exp 2(l+1)*exp 2(l+1)*(1#exp 2(2*(l+1)+k))
(use "RatEqLe" (pt "abs(1/as n*(1/as m))*abs(as m-as n)"))
(use "RatTimesAbs")

; ?_24: abs(1/as n*(1/as m))*abs(as m-as n)<=exp 2(l+1)*exp 2(l+1)*(1#exp 2(2*(l+1)+k))

(add-global-assumption
 "RatAbsTimes"
 (pf "all a,b,a1,b1.abs a<=a1 -> abs b<=b1 -> abs a*abs b<=a1*b1"))

(use "RatAbsTimes")

; ?_25: abs(1/as n*(1/as m))<=exp 2(l+1)*exp 2(l+1)
(use "RatEqLe" (pt "abs(1/as n)*abs(1/as m)"))
(use "RatTimesAbs")
(use "RatAbsTimes")

; ?_29: abs(1/as n)<=exp 2(l+1)
(use "RatInv")
(use "as-pos-l")
(use "M(l+1)<=n")

; ?_30: abs(1/as m)<=exp 2(l+1)
(use "RatInv")
(use "as-pos-l")
(use "M(l+1)<=m")

; ?_26: abs(as m-as n)<=(1#exp 2(2*(l+1)+k))
(use "CauchyElim" (pt "M"))
(auto 1)

; ?_22: exp 2(l+1)*exp 2(l+1)*(1#exp 2(2*(l+1)+k))<=(1#exp 2 k)
(ng)
(ord-field-simp-bwd)

; ?_10: M(l+1)<=m
(use "PosLeTrans" (pt "M(S(S(l+l+k)))"))
(use "MonElim")
(auto 1)
(ng)
(ord-field-simp-bwd)
(assert (pf "S(S(l+l+k))=SZero(S l)+k"))
 (ord-field-simp-bwd)
 (use "Truth-Axiom")
(assume "EqHyp")
(simp "EqHyp")
(auto 1)

; ?_11: M(l+1)<=n
(use "PosLeTrans" (pt "M(S(S(l+l+k)))"))
(use "MonElim")
(auto 1)
(ord-field-simp-bwd)
(assert (pf "S(S(l+l+k))=SZero(S l)+k"))
 (ord-field-simp-bwd)
 (use "Truth-Axiom")
(assume "EqHyp")
(simp "EqHyp")
(auto 1)

(save "CauchyInv")		       

; (dpe)
; (dnpe)
; (cdp)


; We now prove one half of the characterization of positivity for reals,
; to be called "RealPosChar1"

(set-goal
 (pf "allnc as all M,k.Cauchy as M ->
      RealPos(RealConstr as M)k -> 
      ex l,n0 all n.n0<=n -> (1#exp 2 l)<=as n"))
(strip)
(ex-intro (pt "k+1"))
(ex-intro (pt "M(k+1)"))
(strip)

; ?_5: (1#exp 2(k+1))<=as n
(use "RatEqLe" (pt "(IntN 1#exp 2(k+1))+(1#exp 2 k)"))

; ?_6: (1#exp 2(k+1))===((IntN 1#exp 2(k+1))+(1#exp 2 k))
(ord-field-simp-bwd)

; ?_13: (IntN 1#exp 2(k+1))+(1#exp 2 k)<=as n
(use "RatLeTrans" (pt "as n-as(M(k+1))+as(M(k+1))"))
(use "RatPlusLe2")

; ?_10: (IntN 1#exp 2(k+1))<=as n-as(M(k+1))
(add-global-assumption
 "RatAbsLeMinus"
 (pf "all a,k.abs a<=(1#k) -> (IntN 1#k)<=a"))
(use "RatAbsLeMinus")

; ?_12: abs(as n-as(M(k+1)))<=(1#exp 2(k+1))
(use "CauchyElim" (pt "M"))
(auto)

; ?_9: as n-as(M(k+1))+as(M(k+1))<=as n
(ord-field-simp-bwd)

(save "RealPosChar1")

(define eterm (proof-to-extracted-term (theorem-name-to-proof "RealPosChar1")))
(define neterm (nt eterm))
(term-to-expr eterm)
(term-to-expr neterm)
(term-to-string neterm)
; "[M0,n1]S n1@M0(S n1)"


; We prove that every real can be approximated by a rational "RealApprox"

(set-goal (pf "all x,k.Real x -> ex a abs(a-x)<<=1#exp 2 k"))
(ind)
(assume "as" "M" "k" "Real x")
(ex-intro (pt "as(M k)"))
(use "RealLeIntro")
(use "RealNNegIntro")
(assume "l")
(ng)
(cut (pf "abs(as(M k)-as(M(S(S l))))<=(1#exp 2 k)+(1#exp 2 l)"))
(assume "H")
(add-global-assumption
 "RealApproxAux1"
 (pf "all a,b,c.a<=b+c -> 0<=b-a+c"))
(use "RealApproxAux1")
(use "H")

; ?_10: abs(as(M k)-as(M(S(S l))))<=(1#exp 2 k)+(1#exp 2 l)
(use "RatLeTrans"
     (pt "abs(as(M k)-as(M(k+S(S l))))+abs(as(M(k+S(S l)))-as(M(S(S l))))"))
(use "RatAbsLe2")

; ?_14: abs(as(M k)-as(M(k+S(S l))))+abs(as(M(k+S(S l)))-as(M(S(S l))))<=(1#exp 2 k)+(1#exp 2 l)
(use "RatPlusLe2")

; ?_15: abs(as(M k)-as(M(k+S(S l))))<=(1#exp 2 k)
(use "CauchyElim" (pt "M"))
(use "RealElimVariant1")
(use "Real x")
(use "Truth-Axiom")
(use "MonElim")
(use "RealElimVariant2" (pt "as"))
(use "Real x")
(add-global-assumption "RealApproxAux2" (pf "all n,m n<=n+m"))
(use "RealApproxAux2")

; ?_16: abs(as(M(k+S(S l)))-as(M(S(S l))))<=(1#exp 2 l)
(use "CauchyElim" (pt "M"))
(use "RealElimVariant1")
(use "Real x")
(use "MonElim")
(use "RealElimVariant2" (pt "as"))
(use "Real x")
(add-global-assumption "RealApproxAux3" (pf "all n,m n<=m+S(S n)"))
(use "RealApproxAux3")
(use "MonElim")
(use "RealElimVariant2" (pt "as"))
(use "Real x")
(add-global-assumption "RealApproxAux4" (pf "all n n<=S(S n)"))
(use "RealApproxAux4")

(save "RealApprox")

(define RealApprox-eterm
  (proof-to-extracted-term (theorem-name-to-proof "RealApprox")))
(define RealApprox-neterm (nt RealApprox-eterm))
(pp RealApprox-neterm)

"(Rec real=>pos=>rat)([as2,M3,n4]as2(M3 n4))"

(type-to-string (term-to-type RealApprox-neterm))

"real=>pos=>rat"

; Examples of reals

(define real1 (pt "RealConstr([n]1-(1#exp 2 n))([k]k)"))

(pp (nt (mk-term-in-app-form RealApprox-neterm real1 (pt "2"))))

(display-program-constants "PosExp")
(add-computation-rule (pt "exp n(SZero m)") (pt "(exp n m)*(exp n m)"))
(add-computation-rule (pt "exp n(SOne m)") (pt "(exp n(SZero m))*n"))

(pp (nt (mk-term-in-app-form RealApprox-neterm real1 (pt "2"))))
(pp (nt (mk-term-in-app-form RealApprox-neterm real1 (pt "10"))))

"1023#1024"

(remove-computation-rules-for (pt "exp n(SZero m)"))
(remove-computation-rules-for (pt "exp n(SOne m)"))


